/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2010, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert text/UTF based output into XML format.                 *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "xmlout.h"

long record_cntr_printed, record_cntr_blank, record_cntr_unmatched, record_cntr_read, record_cntr_expelled;
char temp_array[TEMP_ARRAY_SIZE];

/* function templates */
void process_input();
void process_output(long record_number);
void print_soap_xmlfy_header();
void print_output_summary(int argc, unsigned char **argv);

void xmlout(argc, argv)
int argc;
unsigned char **argv;
{
  /* process stdin and convert it to XML format */
  int i;
  unsigned char *xsd_filename;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: xmlout%s", newline);

  /* initialise */
  record_cntr_read       = 0;
  record_cntr_blank      = 0;
  record_cntr_unmatched  = 0;
  record_cntr_printed    = 0;
  record_cntr_expelled   = 0;

  /* set the parent XML tag */
  if (sw_schema == SCHEMA_DEFAULT)
  {
    /* schema_parent_name = default_schema_parent_name; */
    /* schema_line_name = default_schema_line_name; */
    schema_parent_name = sw_print[0]->tagname;
    schema_line_name   = sw_print[1]->tagname;
  }
  else
    schema_parent_name = schema_parent.name;
  
  /* pre-load a block of input stream if using BOM option */
  if (sw_bom)
    (void) utf_read_bom(STDIN_FD);
  else 
    utf_getchar_init();

  /* print XML header */
  if (sw_printonly_header)
  {
    if (sw_bom)
      utf_print_bom(STDOUT_FD);

    strcpy_output_record(0, (unsigned char *) "<?xml version=\"");
    if (version_xml == XML_1_0)
      strcat_output_record((unsigned char *) "1.0");
    else if (version_xml == XML_1_1)
      strcat_output_record((unsigned char *) "1.1");

    if (sw_xml_format == SW_XML)
    {
      strcat_output_record((unsigned char *) "\" encoding=\"");
      strcat_output_record((unsigned char *) output_record.encoding_name);
      strcat_output_record((unsigned char *) "\"?>");
      strcat_output_record(newline);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
      if (sw_schema != SCHEMA_XSD)
      {
        strcpy_output_record(0, (unsigned char *) "<!DOCTYPE ");
        strcat_output_record(schema_parent_name);
        strcat_output_record((unsigned char *) " SYSTEM \"");
        if (schema_system_identifier != NULL)
          strcat_output_record(schema_system_identifier);
        else
        {
          if (sw_schema == SCHEMA_DEFAULT)
          {
            strcat_output_record((unsigned char *) progname);
            strcat_output_record((unsigned char *) ".dtd");
          }
          else
            strcat_output_record(schema_filename);
        }
        strcat_output_record((unsigned char *) "\">");
        strcat_output_record(newline);
        utf_print(STDOUT_FD, output_record.line, output_record.length);
      }
    }
    else if (sw_xml_format == SW_SOAP)
    {
      strcat_output_record((unsigned char *) "?>");
      strcat_output_record(newline);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
      strcpy_output_record(0, (unsigned char *) "<");
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_ENVELOPE);
      strcat_output_record((unsigned char *) ":Envelope xmlns:");
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_ENVELOPE);
      strcat_output_record((unsigned char *) "=\"");
      if (version_soap == SOAP_1_1)
      {
        strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_URI_1_1);
        strcat_output_record((unsigned char *) "/soap-envelope\"");
        strcat_output_record(newline);
        strcat_output_record((unsigned char *) "              ");
        strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_ENVELOPE);
        strcat_output_record((unsigned char *) ":encodingStyle=\"");
        strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_URI_1_1);
        strcat_output_record((unsigned char *) "/soap-encoding\">");
        strcat_output_record(newline);
      }
      else if (version_soap == SOAP_1_2)
      {
        strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_URI_1_2);
        strcat_output_record((unsigned char *) "/soap-envelope\">");
        strcat_output_record(newline);
      }
      utf_print(STDOUT_FD, output_record.line, output_record.length);
      /* open SOAP Header */
      strcpy_output_record(1, (unsigned char *) "<");
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_ENVELOPE);
      strcat_output_record((unsigned char *) ":Header>");
      strcat_output_record(newline);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
      /* print SOAP xmlfy header */
      print_soap_xmlfy_header();
      /* close SOAP Header */
      strcpy_output_record(1, (unsigned char *) "</");
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_ENVELOPE);
      strcat_output_record((unsigned char *) ":Header>");
      strcat_output_record(newline);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
      /* open SOAP Body */
      strcpy_output_record(1, (unsigned char *) "<");
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_ENVELOPE);
      strcat_output_record((unsigned char *) ":Body>");
      strcat_output_record(newline);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
    }
  }

  /* open parent tag */
  if (sw_printonly_rtagopen)
  {
    if (sw_xml_format == SW_SOAP)
      strcpy_output_record(2, (unsigned char *) "<");
    else
      strcpy_output_record(0, (unsigned char *) "<");
    if (sw_xml_format == SW_SOAP)
    {
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
      strcat_output_record((unsigned char *) ":");
    }
    /* print the root or parent tag name */
    strcat_output_record(schema_parent_name);
    /* add standard attributes */
    if (sw_xml_format == SW_SOAP)
    {
      strcat_output_record((unsigned char *) " xmlns:");
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
      strcat_output_record((unsigned char *) "=\"");
      if (schema_system_identifier != NULL)
        strcat_output_record(schema_system_identifier);
      else
        strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_URI_XMLFY);
      strcat_output_record((unsigned char *) "\"");
      if (version_soap == SOAP_1_2)
      {
        strcat_output_record(newline);
        strcat_output_record((unsigned char *) "             ");
        strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_ENVELOPE);
        strcat_output_record((unsigned char *) ":encodingStyle=\"");
        strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_URI_1_2);
        strcat_output_record((unsigned char *) "/soap-encoding\"");
      }
    }
    else if (sw_schema == SCHEMA_XSD)
    {
      /* get basename of schema_filename */
      for (i = xstrlen(schema_filename) ; i >= 0 ; i--)
        if (schema_filename[i] == '/' || schema_filename[i] == '\\')
        {
          xsd_filename = &schema_filename[i + 1];
          break;
        }
      if (i < 0)
        xsd_filename = schema_filename;
      
      strcat_output_record((unsigned char *) " ");
      if (schema_system_identifier != NULL)
      {
        strcat_output_record((unsigned char *) "xmlns=\"");
        strcat_output_record(schema_system_identifier);
        strcat_output_record((unsigned char *) "\"");
        strcat_output_record(newline);
        strcat_output_record((unsigned char *) "       ");
      }
      else if (schema_xsd_target_ns != NULL)
      {
        strcat_output_record((unsigned char *) "xmlns=\"");
        strcat_output_record(schema_xsd_target_ns);
        strcat_output_record((unsigned char *) "\"");
        strcat_output_record(newline);
        strcat_output_record((unsigned char *) "       ");
      }
      strcat_output_record((unsigned char *) "xmlns:xsi=\"");
      strcat_output_record((unsigned char *) DEFAULT_STR_XSD_INSTANCE);
      strcat_output_record((unsigned char *) "\"");
      strcat_output_record(newline);
      strcat_output_record((unsigned char *) "       ");
      if (schema_system_identifier != NULL || schema_xsd_target_ns != NULL)
      {
        strcat_output_record((unsigned char *) "xsi:schemaLocation=\"");
        if (schema_system_identifier != NULL)
          strcat_output_record(schema_system_identifier);
        else if (schema_xsd_target_ns != NULL)
          strcat_output_record(schema_xsd_target_ns);
        strcat_output_record((unsigned char *) " ");
        strcat_output_record(xsd_filename);
      }
      else
      {
        strcat_output_record((unsigned char *) "xsi:noNamespaceSchemaLocation=\"");
        strcat_output_record(schema_filename);
      }
      strcat_output_record((unsigned char *) "\"");
    }
    /* add user attributes */
    if (sw_print[0]->timestamp_attr)
    {
      time(&ts);

      (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_TIMESTAMP_DATE);
      strcat_output_record((unsigned char *) temp_array);
      make_timestamp_array(ts, TIMESTAMP_DATE);
      strcat_output_record((unsigned char *) temp_array);
      strcat_output_record((unsigned char *) "\"");

      (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_TIMESTAMP_SECS);
      strcat_output_record((unsigned char *) temp_array);
      make_timestamp_array(ts, TIMESTAMP_SECS);
      strcat_output_record((unsigned char *) temp_array);
      strcat_output_record((unsigned char *) "\"");
    }
    if (nattrname)
    {
      for (i = 0 ; i < nattrname ; i++)
      {
        if (attrname[i].level == 0)
        {
          strcat_output_record((unsigned char *) " ");
          strncat_output_record(attrname[i].name, attrname[i].name_length);
          strcat_output_record((unsigned char *) "=\"");
          strncat_output_record(attrname[i].value, attrname[i].value_length);
          strcat_output_record((unsigned char *) "\"");
        }
      }
    }
    strcat_output_record((unsigned char *) ">");
    strcat_output_record(newline);
    utf_print(STDOUT_FD, output_record.line, output_record.length);
  }

  /* process input */
  if (sw_printonly_records)
    process_input();

  /* close parent tag */
  if (sw_printonly_rtagclose)
  {
    if (sw_xml_format == SW_SOAP)
    {
      strcpy_output_record(2, (unsigned char *) "</");
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
      strcat_output_record((unsigned char *) ":");
    }
    else
      strcpy_output_record(0, (unsigned char *) "</");
    strcat_output_record(schema_parent_name);
    strcat_output_record((unsigned char *) ">");
    strcat_output_record(newline);
    utf_print(STDOUT_FD, output_record.line, output_record.length);
  }

  /* close SOAP Body and Envelope */
  if (sw_printonly_footer)
  {
    if (sw_xml_format == SW_SOAP)
    {
      strcpy_output_record(1, (unsigned char *) "</");
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_ENVELOPE);
      strcat_output_record((unsigned char *) ":Body>");
      strcat_output_record(newline);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
      strcpy_output_record(0, (unsigned char *) "</");
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_ENVELOPE);
      strcat_output_record((unsigned char *) ":Envelope>");
      strcat_output_record(newline);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
    }
  }

  /* print summary */
  if (sw_summary)
  {
    /* summary printed in XML not in SOAP format */
    sw_xml_format = SW_XML;
    print_output_summary(argc, argv);
  }

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: xmlout%s", newline);
}


void print_soap_xmlfy_header()
{
  /* print details of xmlfy in its own SOAP header */

  /* open SOAP xmlfy header */
  strcpy_output_record(2, (unsigned char *) "<");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) ":xmlfyHeader xmlns:");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) "=\"");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_URI_XMLFY_HEADER);
  strcat_output_record((unsigned char *) "\" ");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_ENVELOPE);
  strcat_output_record((unsigned char *) ":mustUnderstand=");
  if (version_soap == SOAP_1_1)
    strcat_output_record((unsigned char *) "\"0\">");
  else if (version_soap == SOAP_1_2)
    strcat_output_record((unsigned char *) "\"false\">");
  strcat_output_record(newline);
  utf_print(STDOUT_FD, output_record.line, output_record.length);
  /* open and close SOAP progname details */
  strcpy_output_record(3, (unsigned char *) "<");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) ":program>");
  strcat_output_record((unsigned char *) progname);
  strcat_output_record((unsigned char *) "</");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) ":program>");
  strcat_output_record(newline);
  utf_print(STDOUT_FD, output_record.line, output_record.length);
  /* open and close SOAP version details */
  strcpy_output_record(3, (unsigned char *) "<");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) ":version>");
  strcat_output_record((unsigned char *) VERSION);
  strcat_output_record((unsigned char *) "</");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) ":version>");
  strcat_output_record(newline);
  utf_print(STDOUT_FD, output_record.line, output_record.length);
  /* open and close SOAP timestamp details */
  strcpy_output_record(3, (unsigned char *) "<");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) ":timestamp>");
  make_timestamp_array(start_ts, TIMESTAMP_SECS);
  strcat_output_record((unsigned char *) temp_array);
  strcat_output_record((unsigned char *) "</");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) ":timestamp>");
  strcat_output_record(newline);
  utf_print(STDOUT_FD, output_record.line, output_record.length);
  /* open and close SOAP user-defined details */
  strcpy_output_record(3, (unsigned char *) "<");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) ":sed>hook</");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) ":sed>");
  strcat_output_record(newline);
  utf_print(STDOUT_FD, output_record.line, output_record.length);
  /* close SOAP xmlfy header */
  strcpy_output_record(2, (unsigned char *) "</");
  strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY_HEADER);
  strcat_output_record((unsigned char *) ":xmlfyHeader>");
  strcat_output_record(newline);
  utf_print(STDOUT_FD, output_record.line, output_record.length);
}
